from typing import TextIO

from .token import Token, TokenType, lookup_hidden, lookup_ident

_IDENT_START = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_"
_IDENT_CONT = _IDENT_START + "0123456789"
_HEXDIGITS = "0123456789abcdefABCDEF"
_HORIZ_WS = " \t\r"
_INDENT_WS = " \t"


class Lexer:
    def __init__(self, textio: TextIO) -> None:
        self._textio = textio
        self._ch = self._read_char()

    def _peek(self, n: int = 1) -> str:
        assert n > 0
        position = self._textio.tell()
        for _ in range(n):
            ch = self._textio.read(1)
        self._textio.seek(position)
        return ch

    def _read_char(self, n: int = 1) -> str:
        for _ in range(n):
            self._ch = self._textio.read(1)
        return self._ch

    def _read_identifier(self) -> str:
        literal = self._ch
        self._read_char()
        while True:
            if self._ch == "":
                break
            if self._ch in _IDENT_CONT:
                literal += self._ch
                self._read_char()
            else:
                break
        return literal

    def _read_number(self) -> str:
        """
        ref: https://www.json.org/json-en.html
        """

        literal = ""
        if self._ch == "-":
            literal += self._ch
            self._read_char()
        if self._ch == "0":
            literal += self._ch
            self._read_char()
        else:
            if self._ch not in "123456789":
                return literal
            literal += self._ch
            self._read_char()
            while self._ch.isdigit():
                literal += self._ch
                self._read_char()
        if self._ch == ".":
            literal += self._ch
            self._read_char()
            if not self._ch.isdigit():
                return literal
            literal += self._ch
            self._read_char()
            while self._ch.isdigit():
                literal += self._ch
                self._read_char()
        if self._ch in ("e", "E"):
            next_ch = self._peek()
            if next_ch in ("+", "-"):
                next_next_ch = self._peek(2)
                if not next_next_ch.isdigit():
                    return literal
                literal += self._ch + next_ch
                self._read_char(2)
            elif next_ch.isdigit():
                literal += self._ch
                self._read_char()
            else:
                return literal
            while self._ch.isdigit():
                literal += self._ch
                self._read_char()
        return literal

    def _read_string(self, verbatim: bool = False) -> str:
        """
        ref: https://www.json.org/json-en.html
        """

        if self._ch not in ('"', "'"):
            raise ValueError(f"unexpected character: {self._ch}")

        quote = self._ch
        self._read_char()

        literal = ""
        while True:
            if self._ch == "":
                raise ValueError("unexpected end of file")
            if self._ch == quote:
                self._read_char()
                break
            elif not verbatim and self._ch == "\\":
                next_ch = self._peek()
                if next_ch in ('"', "'", "\\", "/", "b", "f", "n", "r", "t"):
                    literal += self._ch + self._read_char()
                    self._read_char()
                elif next_ch == "u":
                    if all(self._peek(i + 2) in _HEXDIGITS for i in range(4)):
                        for ch in range(4):
                            literal += self._ch
                            self._read_char()
                    else:
                        raise ValueError(f"unexpected character: {self._ch}")
            else:
                literal += self._ch
                self._read_char()

        return literal

    def _read_text_block(self) -> str:
        if not (self._ch == self._peek(1) == self._peek(2) == "|"):
            raise ValueError(f"unexpected character: {self._ch}")

        # skip the "|||"
        self._read_char(3)

        # chomp whitespace at end of line
        while self._ch in _HORIZ_WS:
            self._read_char()

        if self._ch != "\n":
            raise ValueError("text block syntax requires new line after |||.")

        # skip the new line
        self._read_char()

        literal = ""

        # skip any blank lines at the beginnig of the block
        while self._ch == "\n":
            self._read_char()
            literal += "\n"

        # read the first line and determine the indentation
        indent = ""
        while self._ch in _INDENT_WS:
            indent += self._ch
            self._read_char()
        if not indent:
            raise ValueError("text block's first line must start with whitespace.")
        while self._ch != "\n":
            literal += self._ch
            self._read_char()

        literal += "\n"
        self._read_char()

        while True:
            if self._ch == "":
                raise ValueError("unexpected end of file")
            # check & consume indentation
            for ws in indent:
                if self._ch != ws:
                    if self._ch == self._peek(1) == self._peek(2) == "|":
                        self._read_char(3)
                        break
                    raise ValueError("text block not terminated with |||")
                self._read_char()
            else:
                # read the line
                while self._ch != "\n":
                    if self._ch == "":
                        raise ValueError("unexpected end of file")
                    literal += self._ch
                    self._read_char()
                literal += "\n"
                self._read_char()
                continue
            break

        return literal

    def _read_hidden(self) -> str:
        literal = ""
        for _ in range(3):
            if self._ch == ":":
                literal += self._ch
                self._read_char()
            else:
                break
        return literal

    def _skip_whitespace(self) -> None:
        while self._ch.isspace():
            self._read_char()

    def _skip_single_line_comment(self) -> None:
        while self._ch != "\n" and self._ch != "":
            self._read_char()
        self._skip_whitespace()

    def _skip_block_comment(self) -> None:
        end_found = False
        while not end_found:
            if self._ch == "":
                end_found = True
            elif self._ch == "*" and self._peek() == "/":
                end_found = True
                self._read_char()

            self._read_char()
        self._skip_whitespace()

    def next_token(self) -> Token:
        self._skip_whitespace()

        token: Token

        if self._ch == "=":
            if self._peek() == "=":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.EQEQ, literal)
            else:
                token = Token(TokenType.EQUAL, self._ch)
        elif self._ch == "+":
            token = Token(TokenType.PLUS, self._ch)
        elif self._ch == "-":
            if self._peek().isdigit():
                literal = self._read_number()
                return Token(TokenType.NUMBER, literal)
            token = Token(TokenType.MINUS, self._ch)
        elif self._ch == "*":
            token = Token(TokenType.STAR, self._ch)
        elif self._ch == "/":
            if self._peek() == "/":
                self._skip_single_line_comment()
                return self.next_token()
            elif self._peek() == "*":
                self._skip_block_comment()
                return self.next_token()
            token = Token(TokenType.SLASH, self._ch)
        elif self._ch == "%":
            token = Token(TokenType.MOD, self._ch)
        elif self._ch == "<":
            if self._peek() == "=":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.LE, literal)
            elif self._peek() == "<":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.LSHIFT, literal)
            else:
                token = Token(TokenType.LT, self._ch)
        elif self._ch == ">":
            if self._peek() == "=":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.GE, literal)
            elif self._peek() == ">":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.RSHIFT, literal)
            else:
                token = Token(TokenType.GT, self._ch)
        elif self._ch == "&":
            if self._peek() == "&":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.AND, literal)
            else:
                token = Token(TokenType.LAND, self._ch)
        elif self._ch == "|":
            if self._peek() == "|":
                if self._peek(2) == "|":
                    litera = self._read_text_block()
                    return Token(TokenType.STRING, litera)
                else:
                    ch = self._ch
                    self._read_char()
                    literal = ch + self._ch
                    token = Token(TokenType.OR, literal)
            else:
                token = Token(TokenType.LOR, self._ch)
        elif self._ch == "!":
            if self._peek() == "=":
                ch = self._ch
                self._read_char()
                literal = ch + self._ch
                token = Token(TokenType.NEQ, literal)
            else:
                token = Token(TokenType.NOT, self._ch)
        elif self._ch == "^":
            token = Token(TokenType.XOR, self._ch)
        elif self._ch == ".":
            token = Token(TokenType.DOT, self._ch)
        elif self._ch == ",":
            token = Token(TokenType.COMMA, self._ch)
        elif self._ch == ":":
            literal = self._read_hidden()
            token_type = lookup_hidden(literal)
            return Token(token_type, literal)
        elif self._ch == ";":
            token = Token(TokenType.SEMICOLON, self._ch)
        elif self._ch == "(":
            token = Token(TokenType.LPAREN, self._ch)
        elif self._ch == ")":
            token = Token(TokenType.RPAREN, self._ch)
        elif self._ch == "{":
            token = Token(TokenType.LBRACE, self._ch)
        elif self._ch == "}":
            token = Token(TokenType.RBRACE, self._ch)
        elif self._ch == "[":
            token = Token(TokenType.LBRACKET, self._ch)
        elif self._ch == "]":
            token = Token(TokenType.RBRACKET, self._ch)
        elif self._ch in ('"', "'"):
            literal = self._read_string()
            return Token(TokenType.STRING, literal)
        elif self._ch == "$":
            token = Token(TokenType.DOLLAR, self._ch)
        elif self._ch == "@" and self._peek() in ("'", '"'):
            self._read_char()
            literal = self._read_string(verbatim=True)
            return Token(TokenType.STRING, literal)
        elif self._ch == "#":
            self._skip_single_line_comment()
            return self.next_token()
        elif self._ch == "":
            token = Token(TokenType.EOF, "")
        elif self._ch in _IDENT_START:
            literal = self._read_identifier()
            token_type = lookup_ident(literal)
            return Token(token_type, literal)
        elif self._ch.isdigit():
            return Token(TokenType.NUMBER, self._read_number())
        else:
            token = Token(TokenType.ILLEGAL, self._ch)

        self._read_char()
        return token



import enum
from typing import Any, Callable, Dict, List, Optional, Union

from pysonnet import ast
from pysonnet.lexer import Lexer
from pysonnet.token import TokenType


class Precedence(enum.IntEnum):
    """
    ref: https://jsonnet.org/ref/spec.html#associativity_precedence
    """

    LOWEST = enum.auto()
    OR = enum.auto()
    AND = enum.auto()
    LOR = enum.auto()
    XOR = enum.auto()
    LAND = enum.auto()
    EQUALS = enum.auto()
    LESSGREATER = enum.auto()
    BITSHIFT = enum.auto()
    SUM = enum.auto()
    PRODUCT = enum.auto()
    UNARY = enum.auto()
    INDEX = enum.auto()


_BINARY_PRECEDENCE = {
    TokenType.OR: Precedence.OR,
    TokenType.AND: Precedence.AND,
    TokenType.LOR: Precedence.LOR,
    TokenType.XOR: Precedence.XOR,
    TokenType.LAND: Precedence.LAND,
    TokenType.EQEQ: Precedence.EQUALS,
    TokenType.NEQ: Precedence.EQUALS,
    TokenType.LT: Precedence.LESSGREATER,
    TokenType.GT: Precedence.LESSGREATER,
    TokenType.LE: Precedence.LESSGREATER,
    TokenType.GE: Precedence.LESSGREATER,
    TokenType.IN: Precedence.LESSGREATER,
    TokenType.LSHIFT: Precedence.BITSHIFT,
    TokenType.RSHIFT: Precedence.BITSHIFT,
    TokenType.PLUS: Precedence.SUM,
    TokenType.MINUS: Precedence.SUM,
    TokenType.STAR: Precedence.PRODUCT,
    TokenType.SLASH: Precedence.PRODUCT,
    TokenType.MOD: Precedence.PRODUCT,
    TokenType.LBRACKET: Precedence.INDEX,
    TokenType.LPAREN: Precedence.INDEX,
    TokenType.DOT: Precedence.INDEX,
    TokenType.LBRACE: Precedence.SUM,
}


class Parser:
    def __init__(self, lexer: Lexer) -> None:
        self._lexer = lexer
        self._cur_token = self._lexer.next_token()
        self._peek_token = self._lexer.next_token()

        self._errors: List[str] = []

        self._prefix_parsers: Dict[TokenType, Callable[[], Optional[ast.AST]]] = {
            TokenType.NULL: self._parse_null,
            TokenType.TRUE: self._parse_boolean,
            TokenType.FALSE: self._parse_boolean,
            TokenType.NUMBER: self._parse_number,
            TokenType.STRING: self._parse_string,
            TokenType.IDENT: self._parse_identifier,
            TokenType.LPAREN: self._parse_grouped_expression,
            TokenType.LBRACE: self._parse_object,
            TokenType.LBRACKET: self._parse_array,
            TokenType.LOCAL: self._parse_local_expression,
            TokenType.PLUS: self._parse_unary,
            TokenType.MINUS: self._parse_unary,
            TokenType.BANG: self._parse_unary,
            TokenType.TILDE: self._parse_unary,
            TokenType.SELF: self._parse_self,
            TokenType.DOLLAR: self._parse_dollar,
            TokenType.SUPER: self._parse_super,
            TokenType.FUNCTION: self._parse_function,
            TokenType.ERROR: self._parse_error,
            TokenType.ASSERT: self._parse_assert_expression,
            TokenType.IF: self._parse_if_expression,
            TokenType.IMPORT: self._parse_import,
            TokenType.IMPORTSTR: self._parse_importstr,
            TokenType.IMPORTBIN: self._parse_importbin,
        }
        self._infix_parsers: Dict[TokenType, Callable[[ast.AST], Optional[ast.AST]]] = {
            TokenType.PLUS: self._parse_binary,
            TokenType.MINUS: self._parse_binary,
            TokenType.STAR: self._parse_binary,
            TokenType.SLASH: self._parse_binary,
            TokenType.MOD: self._parse_binary,
            TokenType.LAND: self._parse_binary,
            TokenType.LOR: self._parse_binary,
            TokenType.XOR: self._parse_binary,
            TokenType.LSHIFT: self._parse_binary,
            TokenType.RSHIFT: self._parse_binary,
            TokenType.AND: self._parse_binary,
            TokenType.OR: self._parse_binary,
            TokenType.EQEQ: self._parse_binary,
            TokenType.NEQ: self._parse_binary,
            TokenType.LT: self._parse_binary,
            TokenType.GT: self._parse_binary,
            TokenType.LE: self._parse_binary,
            TokenType.GE: self._parse_binary,
            TokenType.IN: self._parse_binary,
            TokenType.DOT: self._parse_binary,
            TokenType.LBRACKET: self._parse_binary,
            TokenType.LPAREN: self._parse_apply,
            TokenType.LBRACE: self._parse_apply_brace,
        }

    def _peek_error(self, token_type: TokenType) -> None:
        msg = f"expected next token to be {token_type}, got {self._peek_token.token_type} instead"
        self._errors.append(msg)

    def _current_token_type_is(self, token_type: TokenType) -> bool:
        return self._cur_token.token_type == token_type

    def _peek_token_type_is(self, token_type: TokenType) -> bool:
        return self._peek_token.token_type == token_type

    def _expect_peek_type(self, token_type: TokenType) -> bool:
        if self._peek_token_type_is(token_type):
            self.next_token()
            return True
        self._peek_error(token_type)
        return False

    def _current_binary_precedence(self) -> Precedence:
        return _BINARY_PRECEDENCE.get(self._cur_token.token_type, Precedence.LOWEST)

    def _peek_binary_precedence(self) -> Precedence:
        return _BINARY_PRECEDENCE.get(self._peek_token.token_type, Precedence.LOWEST)

    def _parse_null(self) -> ast.Null:
        return ast.Null()

    def _parse_boolean(self) -> ast.Boolean:
        return ast.Boolean(self._cur_token.token_type == TokenType.TRUE)

    def _parse_number(self) -> ast.Number:
        if "." in self._cur_token.literal or "e" in self._cur_token.literal.lower():
            value = float(self._cur_token.literal)
        else:
            value = int(self._cur_token.literal)
        return ast.Number(value=value)

    def _parse_string(self) -> ast.String:
        return ast.String(self._cur_token.literal)

    def _parse_identifier(self) -> ast.Identifier:
        return ast.Identifier(self._cur_token.literal)

    def _parse_grouped_expression(self) -> Optional[ast.AST]:
        self.next_token()  # consume the '(' token
        expression = self._parse_expression(Precedence.LOWEST)
        if not self._expect_peek_type(TokenType.RPAREN):
            return None
        return expression

    def _parse_self(self) -> Optional[ast.Self]:
        return ast.Self()

    def _parse_dollar(self) -> Optional[ast.Dollar]:
        return ast.Dollar()

    def _parse_super(self) -> Optional[ast.Super]:
        return ast.Super()

    def _parse_error(self) -> Optional[ast.Error]:
        self.next_token()  # consume the 'error' token
        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None
        return ast.Error(expression)

    def _parse_assert(self) -> Optional[ast.Assert]:
        self.next_token()  # consume the 'assert' token
        condition = self._parse_expression(Precedence.LOWEST)
        if condition is None:
            return None
        message: Optional[ast.AST] = None
        if self._peek_token_type_is(TokenType.COLON):
            self.next_token()  # consume the ':' token
            self.next_token()  # move to the message
            message = self._parse_expression(Precedence.LOWEST)
            if message is None:
                return None
        return ast.Assert(condition, message)

    def _parse_assert_expression(self) -> Optional[ast.AssertExpression]:
        assert_ = self._parse_assert()
        if not assert_:
            return None
        if not self._expect_peek_type(TokenType.SEMICOLON):
            return None
        self.next_token()
        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None
        return ast.AssertExpression(assert_, expression)

    def _parse_if_expression(self) -> Optional[ast.IfExpression]:
        self.next_token()  # consume the 'if' token
        condition = self._parse_expression(Precedence.LOWEST)
        if not condition:
            return None
        if not self._expect_peek_type(TokenType.THEN):
            return None

        self.next_token()  # consume the 'then' token

        then_expr = self._parse_expression(Precedence.LOWEST)
        if not then_expr:
            return None

        else_expr: Optional[ast.AST] = None
        if self._peek_token_type_is(TokenType.ELSE):
            self.next_token()  # move to the 'else' token
            self.next_token()  # consume the 'else' token
            else_expr = self._parse_expression(Precedence.LOWEST)
            if not else_expr:
                return None

        return ast.IfExpression(condition, then_expr, else_expr)

    def _parse_param(self) -> Optional[ast.Param]:
        ident = self._parse_identifier()
        default: Optional[ast.AST] = None
        if self._peek_token_type_is(TokenType.EQUAL):
            self.next_token()  # move to the '=' token
            self.next_token()  # consume the '=' token
            default = self._parse_expression(Precedence.LOWEST)
            if default is None:
                return None
        return ast.Param(ident, default)

    def _parse_params(self) -> Optional[List[ast.Param]]:
        param = self._parse_param()
        if param is None:
            return None
        params = [param]
        while self._peek_token_type_is(TokenType.COMMA):
            self.next_token()  # consume the ',' token
            if not self._peek_token_type_is(TokenType.IDENT):
                break
            self.next_token()
            param = self._parse_param()
            if param is None:
                return None
            params.append(param)
        return params

    def _parse_function(self) -> Optional[ast.Function]:
        if not self._expect_peek_type(TokenType.LPAREN):
            return None
        self.next_token()
        params = self._parse_params()
        if params is None:
            return None
        if not self._expect_peek_type(TokenType.RPAREN):
            return None
        self.next_token()
        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None
        return ast.Function(params, expression)

    def _parse_apply(self, function: ast.AST) -> Optional[ast.Apply]:
        args: List[ast.Arg] = []
        read_kwargs = False
        while not self._peek_token_type_is(TokenType.RPAREN):
            if read_kwargs:
                if not self._expect_peek_type(TokenType.IDENT):
                    return None
                ident = self._parse_identifier()
                if not self._expect_peek_type(TokenType.EQUAL):
                    return None
                self.next_token()
                expression = self._parse_expression(Precedence.LOWEST)
                if expression is None:
                    return None
                args.append(ast.Arg(expression, ident))
            else:
                self.next_token()
                if self._current_token_type_is(TokenType.IDENT) and self._peek_token_type_is(TokenType.EQUAL):
                    ident = self._parse_identifier()
                    if not self._expect_peek_type(TokenType.EQUAL):
                        return None
                    self.next_token()
                    expression = self._parse_expression(Precedence.LOWEST)
                    if expression is None:
                        return None
                    args.append(ast.Arg(expression, ident))
                else:
                    expression = self._parse_expression(Precedence.LOWEST)
                    if expression is None:
                        return None
                    args.append(ast.Arg(expression))
            if self._peek_token_type_is(TokenType.COMMA):
                self.next_token()
        if not self._expect_peek_type(TokenType.RPAREN):
            return None
        tailstrict = False
        if self._peek_token_type_is(TokenType.TAILSTRICT):
            self.next_token()
            tailstrict = True
        return ast.Apply(function, args, tailstrict)

    def _parse_apply_brace(self, left: ast.AST) -> Optional[ast.ApplyBrace]:
        right = self._parse_object()
        if right is None:
            return None
        return ast.ApplyBrace(left, right)

    def _parse_unary(self) -> Optional[ast.Unary]:
        operator: ast.Unary.Operator
        if self._current_token_type_is(TokenType.PLUS):
            operator = ast.Unary.Operator.PLUS
        elif self._current_token_type_is(TokenType.MINUS):
            operator = ast.Unary.Operator.MINUS
        elif self._current_token_type_is(TokenType.BANG):
            operator = ast.Unary.Operator.NOT
        elif self._current_token_type_is(TokenType.TILDE):
            operator = ast.Unary.Operator.BITWISE_NOT
        else:
            self._errors.append(f"unknown unary operator: {self._cur_token.literal}")
            return None
        self.next_token()
        expression = self._parse_expression(Precedence.UNARY)
        if not expression:
            return None
        return ast.Unary(operator, expression)

    def _parse_binary(self, left: ast.AST) -> Optional[ast.AST]:
        operator: ast.Binary.Operator
        if self._current_token_type_is(TokenType.PLUS):
            operator = ast.Binary.Operator.ADD
        elif self._current_token_type_is(TokenType.MINUS):
            operator = ast.Binary.Operator.SUB
        elif self._current_token_type_is(TokenType.STAR):
            operator = ast.Binary.Operator.MUL
        elif self._current_token_type_is(TokenType.SLASH):
            operator = ast.Binary.Operator.DIV
        elif self._current_token_type_is(TokenType.MOD):
            operator = ast.Binary.Operator.MOD
        elif self._current_token_type_is(TokenType.LAND):
            operator = ast.Binary.Operator.BITWISE_AND
        elif self._current_token_type_is(TokenType.LOR):
            operator = ast.Binary.Operator.BITWISE_OR
        elif self._current_token_type_is(TokenType.XOR):
            operator = ast.Binary.Operator.BITWISE_XOR
        elif self._current_token_type_is(TokenType.LSHIFT):
            operator = ast.Binary.Operator.LSHIFT
        elif self._current_token_type_is(TokenType.RSHIFT):
            operator = ast.Binary.Operator.RSHIFT
        elif self._current_token_type_is(TokenType.AND):
            operator = ast.Binary.Operator.AND
        elif self._current_token_type_is(TokenType.OR):
            operator = ast.Binary.Operator.OR
        elif self._current_token_type_is(TokenType.EQEQ):
            operator = ast.Binary.Operator.EQ
        elif self._current_token_type_is(TokenType.NEQ):
            operator = ast.Binary.Operator.NE
        elif self._current_token_type_is(TokenType.LT):
            operator = ast.Binary.Operator.LT
        elif self._current_token_type_is(TokenType.GT):
            operator = ast.Binary.Operator.GT
        elif self._current_token_type_is(TokenType.LE):
            operator = ast.Binary.Operator.LE
        elif self._current_token_type_is(TokenType.GE):
            operator = ast.Binary.Operator.GE
        elif self._current_token_type_is(TokenType.IN):
            operator = ast.Binary.Operator.IN
        elif self._current_token_type_is(TokenType.DOT):
            if not self._expect_peek_type(TokenType.IDENT):
                return None
            return ast.Binary(
                ast.Binary.Operator.INDEX,
                left,
                ast.String(self._cur_token.literal),
            )
        elif self._current_token_type_is(TokenType.LBRACKET):
            expr = self._parse_index(left)
            if expr is None:
                return None
            return expr
        else:
            self._errors.append(f"unknown binary operator: {self._cur_token.literal}")
            return None
        precedence = self._current_binary_precedence()
        self.next_token()
        right = self._parse_expression(precedence)
        if not right:
            return None
        return ast.Binary(operator, left, right)

    def _parse_index(self, indexable: ast.AST) -> Optional[ast.AST]:
        if self._peek_token_type_is(TokenType.RBRACKET):
            self._errors.append("index expression is empty")
            return None
        start: Optional[ast.AST] = None
        end: Optional[ast.AST] = None
        step: Optional[ast.AST] = None
        num_colons = 0
        if self._peek_token_type_is(TokenType.COLON):
            num_colons += 1
            self.next_token()
        elif self._peek_token_type_is(TokenType.DCOLON):
            num_colons += 2
        else:
            self.next_token()
            start = self._parse_expression(Precedence.LOWEST)
            if start is None:
                return None
            if self._peek_token_type_is(TokenType.COLON):
                num_colons += 1
                self.next_token()
        if self._peek_token_type_is(TokenType.DCOLON):
            num_colons += 2
            self.next_token()
        elif self._peek_token_type_is(TokenType.RBRACKET):
            pass
        else:
            self.next_token()
            end = self._parse_expression(Precedence.LOWEST)
            if not end:
                return None
            if self._peek_token_type_is(TokenType.COLON):
                num_colons += 1
                self.next_token()
        if not self._peek_token_type_is(TokenType.RBRACKET):
            self.next_token()
            step = self._parse_expression(Precedence.LOWEST)
            if not step:
                return None
        if not self._expect_peek_type(TokenType.RBRACKET):
            return None
        if num_colons == 0:
            if start is None:
                return indexable
            return ast.Binary(ast.Binary.Operator.INDEX, indexable, start)
        if not start and not end and not step:
            return indexable
        return ast.Apply(
            ast.Binary(
                ast.Binary.Operator.INDEX,
                ast.Identifier("std"),
                ast.String("slice"),
            ),
            [
                ast.Arg(indexable),
                ast.Arg(start or ast.Null()),
                ast.Arg(end or ast.Null()),
                ast.Arg(step or ast.Null()),
            ],
        )

    def _parse_expression(self, precedence: Precedence) -> Optional[ast.AST]:
        prefix = self._prefix_parsers.get(self._cur_token.token_type)
        if prefix is None:
            self._errors.append(f"no prefix parse function for {self._cur_token.token_type}")
            return None

        left = prefix()
        if left is None:
            return None

        while precedence < self._peek_binary_precedence():
            assert left is not None
            infix = self._infix_parsers.get(self._peek_token.token_type)
            if infix is None:
                return left
            self.next_token()
            left = infix(left)
            if left is None:
                return None

        return left

    def _parse_local_expression(self) -> Optional[ast.LocalExpression]:
        self.next_token()  # consume the 'local' token
        binds: List[ast.Bind] = []

        bind = self._parse_bind()
        if bind is None:
            return None
        binds.append(bind)

        while self._peek_token_type_is(TokenType.COMMA):
            self.next_token()
            if not self._expect_peek_type(TokenType.IDENT):
                return None
            bind = self._parse_bind()
            if bind is None:
                return None
            binds.append(bind)

        if not self._expect_peek_type(TokenType.SEMICOLON):
            return None

        self.next_token()

        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None

        return ast.LocalExpression(binds, expression)

    def _parse_bind(self) -> Optional[ast.Bind]:
        name = ast.Identifier[Any](self._cur_token.literal)
        params: Optional[List[ast.Param]] = None
        if self._peek_token_type_is(TokenType.LPAREN):
            self.next_token()  # move to the '(' token
            if self._peek_token_type_is(TokenType.RPAREN):
                params = []
            else:
                self.next_token()  # consume the '(' token
                params = self._parse_params()
                if not params:
                    return None
            if not self._expect_peek_type(TokenType.RPAREN):
                return None
        if not self._expect_peek_type(TokenType.EQUAL):
            return None
        self.next_token()  # consume the '=' token
        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None
        if params is not None:
            expression = ast.Function(params, expression)
        return ast.Bind(name, expression)

    def _parse_object_local(self) -> Optional[ast.ObjectLocal]:
        self.next_token()  # consume the 'local' token
        bind = self._parse_bind()
        if not bind:
            return None
        return ast.ObjectLocal(bind)

    def _parse_object_field(self) -> Optional[ast.ObjectField]:
        key: ast.AST[str]
        if self._current_token_type_is(TokenType.IDENT):
            key = ast.String(self._cur_token.literal)
        elif self._current_token_type_is(TokenType.STRING):
            key = ast.String(self._cur_token.literal)
        elif self._current_token_type_is(TokenType.LBRACKET):
            self.next_token()  # consume the '[' token
            expression = self._parse_expression(Precedence.LOWEST)
            if expression is None:
                return None
            if not self._expect_peek_type(TokenType.RBRACKET):
                return None
            key = expression
        else:
            self._errors.append("invalid key")
            return None

        params: Optional[List[ast.Param]] = None
        if self._peek_token_type_is(TokenType.LPAREN):
            self.next_token()
            self.next_token()
            params = self._parse_params()
            if params is None:
                return None
            if not self._expect_peek_type(TokenType.RPAREN):
                return None

        inherit = False
        if self._peek_token_type_is(TokenType.PLUS):
            inherit = True
            self.next_token()

        visibility: ast.ObjectField.Visibility
        if self._peek_token_type_is(TokenType.COLON):
            visibility = ast.ObjectField.Visibility.VISIBLE
        elif self._peek_token_type_is(TokenType.DCOLON):
            visibility = ast.ObjectField.Visibility.HIDDEN
        elif self._peek_token_type_is(TokenType.TCOLON):
            visibility = ast.ObjectField.Visibility.FORCE_VISIBLE
        else:
            self._errors.append(f"expected ':' or '::' or ':::', got {self._peek_token.token_type} instead")
            return None

        self.next_token()  # consume the separator token
        self.next_token()

        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None

        if params is not None:
            expression = ast.Function(params, expression)

        return ast.ObjectField(key, expression, inherit, visibility)

    def _parse_object_member(self) -> Optional[ast.ObjectMember]:
        member: Optional[ast.ObjectMember] = None
        if self._current_token_type_is(TokenType.LOCAL):
            member = self._parse_object_local()
        elif self._current_token_type_is(TokenType.ASSERT):
            member = self._parse_assert()
        elif (
            self._current_token_type_is(TokenType.IDENT)
            or self._current_token_type_is(TokenType.STRING)
            or self._current_token_type_is(TokenType.LBRACKET)
        ):
            member = self._parse_object_field()
        else:
            self._errors.append(f"unexpected token {self._cur_token.token_type}")
        if member is None:
            return None
        return member

    def _parse_for_spec(self) -> Optional[ast.ForSpec]:
        self.next_token()  # consume the 'for' token
        identifier = self._parse_identifier()
        if not self._expect_peek_type(TokenType.IN):
            return None
        self.next_token()  # consume the 'in' token
        expression = self._parse_expression(Precedence.LOWEST)
        if expression is None:
            return None
        return ast.ForSpec(identifier, expression)

    def _parse_if_spec(self) -> Optional[ast.IfSpec]:
        self.next_token()  # consume the 'if' token
        condition = self._parse_expression(Precedence.LOWEST)
        if condition is None:
            return None
        return ast.IfSpec(condition)

    def _parse_object(self) -> Optional[Union[ast.Object, ast.ObjectComprehension]]:
        fields: List[ast.ObjectField] = []
        efields: List[ast.ObjectField] = []
        asserts: List[ast.Assert] = []
        locals_: List[ast.ObjectLocal] = []
        members: List[ast.ObjectMember] = []
        while not (self._peek_token_type_is(TokenType.RBRACE) or self._peek_token_type_is(TokenType.FOR)):
            container: List
            if self._peek_token_type_is(TokenType.LOCAL):
                container = locals_
            elif self._peek_token_type_is(TokenType.ASSERT):
                container = asserts
            elif self._peek_token_type_is(TokenType.LBRACKET):
                container = efields
            else:
                container = fields
            self.next_token()
            member = self._parse_object_member()
            if member is None:
                return None
            container.append(member)
            members.append(member)
            if self._peek_token_type_is(TokenType.COMMA):
                self.next_token()
        if self._peek_token_type_is(TokenType.FOR):
            if fields or len(efields) != 1:
                self._errors.append("object comprehensions can only have a single [e] field")
                return None
            key = efields[0].key
            value = efields[0].expr
            self.next_token()  # move to the 'for' token
            forspec = self._parse_for_spec()
            if forspec is None:
                return None
            compspecs: List[ast.ComprehensionSpec] = []
            while not self._peek_token_type_is(TokenType.RBRACE):
                if self._peek_token_type_is(TokenType.FOR):
                    self.next_token()
                    nested_forspec = self._parse_for_spec()
                    if nested_forspec is None:
                        return None
                    compspecs.append(nested_forspec)
                elif self._peek_token_type_is(TokenType.IF):
                    self.next_token()
                    nested_ifspec = self._parse_if_spec()
                    if nested_ifspec is None:
                        return None
                    compspecs.append(nested_ifspec)
                else:
                    self._errors.append(f"expected 'for' or 'if', got {self._peek_token.token_type} instead")
                    return None
            if not self._expect_peek_type(TokenType.RBRACE):
                return None
            return ast.ObjectComprehension(locals_, key, value, forspec, compspecs)
        if not self._expect_peek_type(TokenType.RBRACE):
            return None
        return ast.Object(members)

    def _parse_array(self) -> Optional[Union[ast.Array, ast.ArrayComprehension]]:
        if self._peek_token_type_is(TokenType.RBRACKET):
            self.next_token()
            return ast.Array([])

        self.next_token()  # consume the '[' token
        first_expression = self._parse_expression(Precedence.LOWEST)
        if first_expression is None:
            return None

        if self._peek_token_type_is(TokenType.COMMA):
            self.next_token()

        # parse array comprehension
        if self._peek_token_type_is(TokenType.FOR):
            self.next_token()  # move to the 'for' token
            forspec = self._parse_for_spec()
            if forspec is None:
                return None
            compspec: List[ast.ComprehensionSpec] = []
            while not self._peek_token_type_is(TokenType.RBRACKET):
                if self._peek_token_type_is(TokenType.FOR):
                    self.next_token()
                    nested_forspec = self._parse_for_spec()
                    if nested_forspec is None:
                        return None
                    compspec.append(nested_forspec)
                elif self._peek_token_type_is(TokenType.IF):
                    self.next_token()
                    nested_ifspec = self._parse_if_spec()
                    if nested_ifspec is None:
                        return None
                    compspec.append(nested_ifspec)
                else:
                    self._errors.append(f"expected 'for' or 'if', got {self._peek_token.token_type} instead")
                    return None
            if not self._expect_peek_type(TokenType.RBRACKET):
                return None
            return ast.ArrayComprehension(first_expression, forspec, compspec)

        # parse array
        elements: List[ast.AST[Any]] = [first_expression]
        if self._peek_token_type_is(TokenType.COMMA):
            self.next_token()  # consume the ',' token
        while not self._peek_token_type_is(TokenType.RBRACKET):
            self.next_token()
            element = self._parse_expression(Precedence.LOWEST)
            if element is None:
                return None
            elements.append(element)
            if self._peek_token_type_is(TokenType.COMMA):
                self.next_token()
        if not self._expect_peek_type(TokenType.RBRACKET):
            return None
        return ast.Array(elements)

    def _parse_import(self) -> Optional[ast.Import]:
        if not self._expect_peek_type(TokenType.STRING):
            return None
        filename = self._cur_token.literal
        return ast.Import(filename)

    def _parse_importstr(self) -> Optional[ast.Importstr]:
        if not self._expect_peek_type(TokenType.STRING):
            return None
        filename = self._cur_token.literal
        return ast.Importstr(filename)

    def _parse_importbin(self) -> Optional[ast.Importbin]:
        if not self._expect_peek_type(TokenType.STRING):
            return None
        filename = self._cur_token.literal
        return ast.Importbin(filename)

    @property
    def errors(self) -> List[str]:
        return self._errors

    def next_token(self) -> None:
        self._cur_token = self._peek_token
        self._peek_token = self._lexer.next_token()

    def parse(self) -> Optional[ast.AST]:
        return self._parse_expression(Precedence.LOWEST)



import dataclasses
from enum import Enum


class TokenType(Enum):
    ILLEGAL = "ILLEGAL"
    EOF = "EOF"

    # Identifiers + literals
    IDENT = "IDENT"
    STRING = "STRING"
    NUMBER = "NUMBER"

    # Operators
    EQUAL = "="
    PLUS = "+"
    MINUS = "-"
    STAR = "*"
    SLASH = "/"
    MOD = "%"
    LT = "<"
    GT = ">"
    EQEQ = "=="
    NEQ = "!="
    LE = "<="
    GE = ">="
    AND = "&&"
    OR = "||"
    NOT = "!"
    LAND = "&"
    LOR = "|"
    XOR = "^"
    LSHIFT = "<<"
    RSHIFT = ">>"
    BANG = "!"
    TILDE = "~"

    # Delimiters
    DOT = "."
    COMMA = ","
    COLON = ":"
    DCOLON = "::"
    TCOLON = ":::"
    SEMICOLON = ";"
    LPAREN = "("
    RPAREN = ")"
    LBRACE = "{"
    RBRACE = "}"
    LBRACKET = "["
    RBRACKET = "]"

    # Keywords
    LOCAL = "LOCAL"
    NULL = "NULL"
    TRUE = "TRUE"
    FALSE = "FALSE"
    FUNCTION = "FUNCTION"
    SELF = "SELF"
    SUPER = "SUPER"
    IF = "IF"
    THEN = "THEN"
    ELSE = "ELSE"
    FOR = "FOR"
    IN = "IN"
    ASSERT = "ASSERT"
    ERROR = "ERROR"
    IMPORT = "IMPORT"
    IMPORTSTR = "IMPORTSTR"
    IMPORTBIN = "IMPORTBIN"
    DOLLAR = "DOLLAR"
    TAILSTRICT = "TAILSTRICT"


_KEYWORDS = {
    "local": TokenType.LOCAL,
    "null": TokenType.NULL,
    "true": TokenType.TRUE,
    "false": TokenType.FALSE,
    "function": TokenType.FUNCTION,
    "self": TokenType.SELF,
    "super": TokenType.SUPER,
    "if": TokenType.IF,
    "then": TokenType.THEN,
    "else": TokenType.ELSE,
    "for": TokenType.FOR,
    "in": TokenType.IN,
    "assert": TokenType.ASSERT,
    "error": TokenType.ERROR,
    "import": TokenType.IMPORT,
    "importstr": TokenType.IMPORTSTR,
    "importbin": TokenType.IMPORTBIN,
    "tailstrict": TokenType.TAILSTRICT,
}

_HIDDEN = {
    ":": TokenType.COLON,
    "::": TokenType.DCOLON,
    ":::": TokenType.TCOLON,
}


def lookup_ident(ident: str) -> TokenType:
    return _KEYWORDS.get(ident, TokenType.IDENT)


def lookup_hidden(hidden: str) -> TokenType:
    return _HIDDEN.get(hidden, TokenType.ILLEGAL)


@dataclasses.dataclass(frozen=True)
class Token:
    token_type: TokenType
    literal: str



from __future__ import annotations

import copy
import enum
import json
from typing import Callable, Dict, Generic, List, NamedTuple, Optional, TypeVar, Union, overload

from pysonnet.types import JsonPrimitive

_Self = TypeVar("_Self", bound="Primitive")
_Real_co = TypeVar("_Real_co", bound=Union[bool, int, float], covariant=True)
_Real_contra = TypeVar("_Real_contra", bound=Union[bool, int, float], contravariant=True)
_PrimitiveType = TypeVar("_PrimitiveType", bound="Primitive")


class Primitive:
    def __str__(self) -> str:
        return json.dumps(self.to_json())

    def __repr__(self) -> str:
        return self.__str__()

    def to_json(self) -> JsonPrimitive:
        """Convert the object to a JSON-serializable object."" """
        raise NotImplementedError

    def clone(self: _Self) -> _Self:
        return copy.deepcopy(self)


class Lazy(Primitive):
    def __init__(self, constructor: Callable[[], Primitive]) -> None:
        self._constructor = constructor

    def __call__(self) -> Primitive:
        value = self._constructor()
        while isinstance(value, Lazy):
            value = value()
        return value

    def __str__(self) -> str:
        return json.dumps(self.to_json())

    def __repr__(self) -> str:
        return f"Lazy({self._constructor})"

    def to_json(self) -> JsonPrimitive:
        return self().to_json()

    def clone(self) -> "Lazy":
        return self


class Null(Primitive):
    def to_json(self) -> None:
        return None


class Number(Generic[_Real_co], Primitive):
    def __init__(self, value: _Real_co) -> None:
        self.value: _Real_co = value

    @overload
    def __add__(self: Number[Union[bool, int]], other: Number[Union[bool, int]]) -> Number[int]:
        ...

    @overload
    def __add__(self: Number[Union[bool, int]], other: Number[float]) -> Number[float]:
        ...

    @overload
    def __add__(self: Number[float], other: Number[Union[bool, int]]) -> Number[float]:
        ...

    @overload
    def __add__(self: Number[float], other: Number[float]) -> Number[float]:
        ...

    @overload
    def __add__(self, other: String) -> String:
        ...

    def __add__(
        self, other: Union[Number[Union[bool, int, float]], String]
    ) -> Union[Number[Union[int, float]], String]:
        if isinstance(other, String):
            return String(json.dumps(self.to_json())) + other
        return Number(self.value + other.value)

    @overload
    def __sub__(self: Number[Union[bool, int]], other: Number[Union[bool, int]]) -> Number[int]:
        ...

    @overload
    def __sub__(self: Number[Union[bool, int]], other: Number[float]) -> Number[float]:
        ...

    @overload
    def __sub__(self: Number[float], other: Number[Union[bool, int]]) -> Number[float]:
        ...

    @overload
    def __sub__(self: Number[float], other: Number[float]) -> Number[float]:
        ...

    def __sub__(self, other: Number[Union[bool, int, float]]) -> Number[Union[int, float]]:
        return Number(self.value - other.value)

    @overload
    def __mul__(self: Number[Union[bool, int]], other: Number[Union[bool, int]]) -> Number[int]:
        ...

    @overload
    def __mul__(self: Number[Union[bool, int]], other: Number[float]) -> Number[float]:
        ...

    @overload
    def __mul__(self: Number[float], other: Number[Union[bool, int]]) -> Number[float]:
        ...

    @overload
    def __mul__(self: Number[float], other: Number[float]) -> Number[float]:
        ...

    def __mul__(self, other: Number[Union[bool, int, float]]) -> Number[Union[int, float]]:
        return Number(self.value * other.value)

    @overload
    def __truediv__(self: Number[Union[bool, int]], other: Number[Union[bool, int]]) -> Number[float]:
        ...

    @overload
    def __truediv__(self: Number[Union[bool, int]], other: Number[float]) -> Number[float]:
        ...

    @overload
    def __truediv__(self: Number[float], other: Number[Union[bool, int]]) -> Number[float]:
        ...

    @overload
    def __truediv__(self: Number[float], other: Number[float]) -> Number[float]:
        ...

    def __truediv__(self, other: Number[Union[bool, int, float]]) -> Number[float]:
        return Number(self.value / other.value)

    @overload
    def __mod__(self: Number[Union[bool, int]], other: Number[Union[bool, int]]) -> Number[int]:
        ...

    @overload
    def __mod__(self: Number[Union[bool, int]], other: Number[float]) -> Number[float]:
        ...

    @overload
    def __mod__(self: Number[float], other: Number[Union[bool, int]]) -> Number[float]:
        ...

    @overload
    def __mod__(self: Number[float], other: Number[float]) -> Number[float]:
        ...

    def __mod__(self, other: Number[Union[bool, int, float]]) -> Number[Union[int, float]]:
        return Number(self.value % other.value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Number):
            return NotImplemented
        return bool(self.value == other.value)

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Number):
            return NotImplemented
        return bool(self.value != other.value)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Number):
            return NotImplemented
        return bool(self.value < other.value)

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Number):
            return NotImplemented
        return bool(self.value > other.value)

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Number):
            return NotImplemented
        return bool(self.value <= other.value)

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Number):
            return NotImplemented
        return bool(self.value >= other.value)

    def __abs__(self) -> Number[Union[int, float]]:
        return Number(abs(self.value))

    def __neg__(self) -> Number[Union[int, float]]:
        return Number(-self.value)

    def __bool__(self) -> bool:
        return bool(self.value)

    def __int__(self) -> int:
        return int(self.value)

    def __float__(self) -> float:
        return float(self.value)

    def to_json(self) -> Union[bool, int, float]:
        return self.value


class Boolean(Number[bool]):
    ...


class String(str, Primitive):
    def __add__(self, other: Union[Number, String, Array, Object]) -> String:  # type: ignore[override]
        if not isinstance(other, String):
            other = String(json.dumps(other.to_json()))
        return String(super().__add__(other))

    def __radd__(self, other: Union[Number, String, Array, Object]) -> String:  # type: ignore[override]
        if not isinstance(other, String):
            other = String(json.dumps(other.to_json()))
        return other + self

    def __mod__(self, other: Primitive) -> String:
        if isinstance(other, Lazy):
            other = other()
        if isinstance(other, Number):
            other = other.to_json()  # type: ignore[assignment]
        if isinstance(other, Array):
            other = tuple(other)  # type: ignore[assignment]
        return String(super().__mod__(other))

    def to_json(self) -> str:
        return self


class Array(Generic[_PrimitiveType], List[_PrimitiveType], Primitive):
    def __add__(self, other: Array) -> Array:  # type: ignore[override]
        return Array(super().__add__(other))

    def to_json(self) -> List[JsonPrimitive]:
        return [item.to_json() for item in self]


class Object(Dict[String, Primitive], Primitive):
    class Visibility(enum.Enum):
        VISIBLE = enum.auto()
        HIDDEN = enum.auto()
        FORCE_VISIBLE = enum.auto()

    class Field(NamedTuple):
        key: String
        value: Primitive
        inherit: bool = False
        visibility: Optional[Object.Visibility] = None

    def __init__(self, *fields: Field) -> None:
        super().__init__(**{field.key: field.value for field in fields})
        self._properties = {
            field.key: (field.inherit, field.visibility or Object.Visibility.VISIBLE) for field in fields
        }

    def add_field(self, field: Field) -> None:
        key = field.key
        value = field.value
        inherit = field.inherit
        visibility = field.visibility or Object.Visibility.VISIBLE
        if key in self:
            if inherit:
                current_value = self[key]

                def _add(left: Primitive = current_value, right: Primitive = value) -> Primitive:
                    if isinstance(left, Lazy):
                        left = left()
                    if isinstance(right, Lazy):
                        right = right()
                    return left + right  # type: ignore[operator, no-any-return]

                value = Lazy(_add)
            if self.hidden(key) and visibility != Object.Visibility.FORCE_VISIBLE:
                visibility = Object.Visibility.HIDDEN
            inherit = self.inherit(key) and inherit
        self[key] = value
        self._properties[key] = (inherit, visibility)

    def get_field(self, key: String) -> Optional[Field]:
        if key in self:
            return Object.Field(key, self[key], *self._properties[key])
        return None

    def inherit(self, key: String) -> bool:
        return self._properties[key][0]

    def visibility(self, key: String) -> Object.Visibility:
        return self._properties[key][1]

    def hidden(self, key: String) -> bool:
        return self.visibility(key) == Object.Visibility.HIDDEN

    @property
    def fields(self) -> List[Field]:
        return [Object.Field(key, value, self.inherit(key), self.visibility(key)) for key, value in self.items()]

    @overload
    def __add__(self, other: Object) -> Object:  # type: ignore[override]
        ...

    @overload
    def __add__(self, other: String) -> String:  # type: ignore[override]
        ...

    def __add__(self, other: Union[Object, String]) -> Union[Object, String]:  # type: ignore[override]
        if isinstance(other, String):
            return String(json.dumps(self.to_json())) + other
        obj = self
        for field in other.fields:
            obj.add_field(field)
        return obj

    def to_json(self) -> Dict[str, JsonPrimitive]:
        return {key: value.to_json() for key, value in self.items() if self.visibility(key) != Object.Visibility.HIDDEN}


class Function(Generic[_PrimitiveType], Primitive):
    def __init__(self, func: Callable[..., _PrimitiveType]) -> None:
        self._func = func

    def __call__(self, *args: Primitive, **kwargs: Primitive) -> _PrimitiveType:
        return self._func(*args, **kwargs)

    def __str__(self) -> str:
        raise NotImplementedError

    def to_json(self) -> None:
        raise NotImplementedError


# Constants
NULL = Null()
TRUE = Boolean(True)
FALSE = Boolean(False)



import inspect
import sys
from functools import wraps
from typing import Any, Callable, Mapping, TypeVar, Union

from pysonnet.errors import PysonnetRuntimeError
from pysonnet.objects import NULL, TRUE, Array, Boolean, Function, Lazy, Null, Number, Object, Primitive, String

_T = TypeVar("_T", bound=Primitive)
_T_co = TypeVar("_T_co", covariant=True, bound=Primitive)
_T_contra = TypeVar("_T_contra", contravariant=True, bound=Primitive)


def _eval_args(func: Callable[..., _T]) -> Callable[..., _T]:
    signature = inspect.signature(func)

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> _T:
        # TODO: check types
        try:
            bound_args = signature.bind(*args, **kwargs)
            bound_args.apply_defaults()
            params = bound_args.arguments
        except TypeError:
            raise PysonnetRuntimeError(f"Invalid arguments of {func.__name__}")
        evaluated_params = {k: v() if isinstance(v, Lazy) else v for k, v in params.items()}
        return func(**evaluated_params)

    return wrapper


class StdLib:
    def __init__(self, ext_vars: Mapping[str, str] = {}) -> None:
        self._ext_vars = ext_vars

    @_eval_args
    def _ext_var(self, x: String) -> String:
        if x not in self._ext_vars:
            raise PysonnetRuntimeError(f"Undefined external variable: {x}")
        return String(self._ext_vars[x])

    @_eval_args
    def _type(self, x: Primitive) -> String:
        if isinstance(x, Null):
            return String("null")
        if isinstance(x, Boolean):
            return String("boolean")
        if isinstance(x, Number):
            return String("number")
        if isinstance(x, String):
            return String("string")
        if isinstance(x, Array):
            return String("array")
        if isinstance(x, Object):
            return String("object")
        if isinstance(x, Function):
            return String("function")
        raise TypeError(type(x))

    @_eval_args
    def _length(self, x: Primitive) -> Number[int]:
        if isinstance(x, String):
            return Number(len(x))
        if isinstance(x, Array):
            return Number(len(x))
        if isinstance(x, Object):
            return Number(len(x))
        raise PysonnetRuntimeError(f"Cannot get length of {self._type(x)}")

    @_eval_args
    def _get(
        self,
        o: Object,
        f: String,
        default: Primitive = NULL,
        inc_hidden: Boolean = TRUE,
    ) -> Primitive:
        value = o.get(f)
        if value is None or (not inc_hidden and o.hidden(f)):
            return default
        return value

    @_eval_args
    def _object_has(self, o: Object, f: String) -> Boolean:
        return Boolean(f in o and not o.hidden(f))

    @_eval_args
    def _slice(
        self,
        iterable: Array[_T],
        start: Union[Number[int], Null],
        end: Union[Number[int], Null],
        step: Union[Number[int], Null],
    ) -> Array[_T]:
        s = slice(
            start.value if isinstance(start, Number) else None,
            end.value if isinstance(end, Number) else None,
            step.value if isinstance(step, Number) else None,
        )
        return Array(iterable[s])

    @_eval_args
    def _range(
        self,
        start: Number[int],
        end: Number[int],
        step: Union[Number[int], Null] = NULL,
    ) -> Array[Number[int]]:
        return Array(
            [
                Number(i)
                for i in range(
                    start.value,
                    end.value + 1,
                    step.value if isinstance(step, Number) else 1,
                )
            ]
        )

    @_eval_args
    def _filter(self, func: Function, value: Array[_T]) -> Array[_T]:
        return Array([item for item in value if func(item)])

    @_eval_args
    def _map(self, func: Function, value: Primitive) -> Primitive:
        if isinstance(value, Array):
            return Array(map(func, value))
        raise PysonnetRuntimeError(f"Cannot map over {self._type(value)}")

    @_eval_args
    def _make_array(self, size: int, func: Function[_T]) -> Array[_T]:
        return Array([func(Number(i)) for i in range(size)])

    @_eval_args
    def _join(self, delimiter: String, value: Array[String]) -> String:
        return String(delimiter.join(value))

    @_eval_args
    def _codepoint(self, value: String) -> Number[int]:
        return Number(ord(value))

    @_eval_args
    def _abs(self, value: Number) -> Number:
        return abs(Number(value))  # type: ignore[type-var]

    @_eval_args
    def _max(self, a: Number, b: Number) -> Number:
        return Number(max(a, b))

    @_eval_args
    def _min(self, a: Number, b: Number) -> Number:
        return Number(min(a, b))

    @_eval_args
    def _clamp(self, a: Number, minVal: Number, maxVal: Number) -> Number:
        return Number(max(min(a, maxVal), minVal))

    @_eval_args
    def _to_string(self, a: Primitive) -> String:
        return String(a)

    @_eval_args
    def _format(self, str: String, vals: Union[Null, Number, String, Array, Object]) -> String:
        return str % vals

    @_eval_args
    def _trace(self, str: String, rest: _T) -> _T:
        print("TRACE:", str, file=sys.stderr)
        return rest

    def as_object(self) -> Object:
        return Object(
            Object.Field(String("extVar"), Function(self._ext_var)),
            Object.Field(String("type"), Function(self._type)),
            Object.Field(String("length"), Function(self._length)),
            Object.Field(String("get"), Function(self._get)),
            Object.Field(String("objectHas"), Function(self._object_has)),
            Object.Field(String("slice"), Function(self._slice)),
            Object.Field(String("range"), Function(self._range)),
            Object.Field(String("map"), Function(self._map)),
            Object.Field(String("makeArray"), Function(self._make_array)),
            Object.Field(String("join"), Function(self._join)),
            Object.Field(String("codepoint"), Function(self._codepoint)),
            Object.Field(String("abs"), Function(self._abs)),
            Object.Field(String("max"), Function(self._max)),
            Object.Field(String("min"), Function(self._min)),
            Object.Field(String("clamp"), Function(self._clamp)),
            Object.Field(String("toString"), Function(self._to_string)),
            Object.Field(String("format"), Function(self._format)),
            Object.Field(String("trace"), Function(self._trace)),
        )



import dataclasses
from os import PathLike
from pathlib import Path
from typing import Dict, List, Mapping, Optional, Union, cast

from pysonnet import ast
from pysonnet.errors import PysonnetRuntimeError
from pysonnet.lexer import Lexer
from pysonnet.objects import FALSE, NULL, TRUE, Array, Boolean, Function, Lazy, Null, Number, Object, Primitive, String
from pysonnet.parser import Parser
from pysonnet.stdlib import StdLib


def _make_lazy(evaluator: "Evaluator", node: ast.AST, context: "Context") -> Lazy:
    # specify default args to capture the values
    return Lazy(lambda evalualtor=evaluator, node=node, context=context: evaluator(node, context))  # type: ignore[misc]


@dataclasses.dataclass
class Context:
    bindings: Dict[str, Union[Lazy, Primitive]] = dataclasses.field(default_factory=dict)
    dollar: Optional[Object] = None
    super_: Optional[Object] = None
    this: Optional[Object] = None

    def clone(self) -> "Context":
        return Context(
            {k: v.clone() for k, v in self.bindings.items()},
            self.dollar.clone() if self.dollar is not None else None,
            self.super_.clone() if self.super_ is not None else None,
            self.this.clone() if self.this is not None else None,
        )


class Evaluator:
    def __init__(
        self,
        ext_vars: Optional[Mapping[str, str]] = None,
        filename: Optional[Union[str, PathLike]] = None,
    ) -> None:
        ext_vars = ext_vars or {}
        filename = Path(filename) if filename else None
        self._ext_vars = ext_vars
        self._filename = filename
        self._stdlib = StdLib(ext_vars)
        self._stdobj = self._stdlib.as_object()
        self._rootdir = self._filename.parent if self._filename else Path("")
        self._stdobj.add_field(Object.Field(String("thisFile"), String(self._filename or "")))

    def _evaluate_literal(self, node: ast.LiteralAST, context: Context) -> Primitive:
        del context
        if isinstance(node, ast.Null):
            return NULL
        if isinstance(node, ast.Boolean):
            return TRUE if node.value else FALSE
        if isinstance(node, ast.Number):
            return Number(node.value)
        if isinstance(node, ast.String):
            return String(node.value)
        raise PysonnetRuntimeError(f"Unsupported literal type: {type(node)}")

    def _evaluate_identifier(self, node: ast.Identifier, context: Context) -> Primitive:
        if node.name in context.bindings:
            value = context.bindings[node.name]
            if isinstance(value, Lazy):
                value = value()
            return value
        if node.name == "std":
            return self._stdobj
        raise PysonnetRuntimeError(f"Unknown variable: {node.name}")

    def _evaluate_object(self, node: ast.Object, context: Context) -> Object:
        bindings = {}
        objfields: Dict[String, ast.ObjectField] = {}
        for member in node.members:
            if isinstance(member, ast.ObjectField):
                key = self(member.key, context)
                if isinstance(key, Null):
                    continue
                if not isinstance(key, String):
                    raise PysonnetRuntimeError(f"Field name must be a string, not {self._stdlib._type(key)}")
                if key in objfields:
                    raise PysonnetRuntimeError(f"Duplicate field: {key}")
                objfields[key] = member
            elif isinstance(member, ast.ObjectLocal):
                bindings[member.bind.ident.name] = _make_lazy(self, member.bind.expr, context)
            elif isinstance(member, ast.Assert):
                condition = self(member.condition, context)
                if not isinstance(condition, Boolean):
                    raise PysonnetRuntimeError(f"Unpexpected type {self._stdlib._type(condition)}, expected boolean")
                if not condition:
                    message: Primitive
                    if member.message is not None:
                        message = self(member.message, context)
                    else:
                        message = String("Object assertion failed")
                    raise PysonnetRuntimeError(str(message))
                continue
            else:
                raise PysonnetRuntimeError(f"Unsupported object member: {type(member)}")
        obj = Object()
        context = context.clone()
        context.bindings.update(bindings)
        context.this = obj
        if context.dollar is None:
            context.dollar = obj
        for key, field in objfields.items():
            value = _make_lazy(self, field.expr, context)
            inherit = field.inherit
            visibility = Object.Visibility.VISIBLE
            if field.visibility == ast.ObjectField.Visibility.HIDDEN:
                visibility = Object.Visibility.HIDDEN
            elif field.visibility == ast.ObjectField.Visibility.FORCE_VISIBLE:
                visibility = Object.Visibility.FORCE_VISIBLE
            obj.add_field(Object.Field(key, value, inherit, visibility))
        return obj

    def _evaluate_array(self, node: ast.Array, context: Context) -> Array:
        return Array([self(expr, context) for expr in node.elements])

    def _evaluate_binary(self, node: ast.Binary, context: Context) -> Primitive:
        operator = node.operator
        left = self(node.left, context)
        if isinstance(left, Object):
            context = context.clone()
            context.super_ = left
        right = self(node.right, context)
        if isinstance(left, Lazy):
            left = left()
        if isinstance(right, Lazy):
            right = right()
        del context
        if operator == ast.Binary.Operator.ADD:
            if isinstance(left, Number) and isinstance(right, Number):
                return left + right
            if isinstance(left, Array) and isinstance(right, Array):
                return left + right
            if isinstance(left, String) and isinstance(right, (Number, String, Array, Object)):
                return left + right
            if isinstance(left, (Number, String, Array, Object)) and isinstance(right, String):
                return left + right
            if isinstance(left, Object) and isinstance(right, Object):
                # TODO: handle errors
                return left + right
            raise PysonnetRuntimeError(
                f"Unsupported operand types for +: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.SUB:
            if isinstance(left, Number) and isinstance(right, Number):
                return left - right
            raise PysonnetRuntimeError(
                f"Unsupported operand types for -: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.MUL:
            if isinstance(left, Number) and isinstance(right, Number):
                return left * right
            raise PysonnetRuntimeError(
                f"Unsupported operand types for *: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.DIV:
            if isinstance(left, Number) and isinstance(right, Number):
                try:
                    return left / right
                except ZeroDivisionError:
                    raise PysonnetRuntimeError("Division by zero")
            raise PysonnetRuntimeError(
                f"Unsupported operand types for /: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.MOD:
            if isinstance(left, Number) and isinstance(right, Number):
                return left % right
            if isinstance(left, String) and isinstance(right, (Number, String, Array, Object)):
                try:
                    return left % right
                except TypeError as e:
                    raise PysonnetRuntimeError(e.args[0])
            raise PysonnetRuntimeError(
                f"Unsupported operand types for %: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.EQ:
            return Boolean(left == right)
        if operator == ast.Binary.Operator.NE:
            return Boolean(left != right)
        if operator == ast.Binary.Operator.LT:
            if isinstance(left, Number) and isinstance(right, Number):
                return Boolean(left < right)
            if isinstance(left, String) and isinstance(right, String):
                return Boolean(left < right)
            if isinstance(left, Array) and isinstance(right, Array):
                return Boolean(left < right)
            raise PysonnetRuntimeError(
                f"Unsupported operand types for <: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.LE:
            if isinstance(left, Number) and isinstance(right, Number):
                return Boolean(left <= right)
            if isinstance(left, String) and isinstance(right, String):
                return Boolean(left <= right)
            if isinstance(left, Array) and isinstance(right, Array):
                return Boolean(left <= right)
            raise PysonnetRuntimeError(
                f"Unsupported operand types for <=: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.GT:
            if isinstance(left, Number) and isinstance(right, Number):
                return Boolean(left > right)
            if isinstance(left, String) and isinstance(right, String):
                return Boolean(left > right)
            if isinstance(left, Array) and isinstance(right, Array):
                return Boolean(left > right)
            raise PysonnetRuntimeError(
                f"Unsupported operand types for >: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.GE:
            if isinstance(left, Number) and isinstance(right, Number):
                return Boolean(left >= right)
            if isinstance(left, String) and isinstance(right, String):
                return Boolean(left >= right)
            if isinstance(left, Array) and isinstance(right, Array):
                return Boolean(left >= right)
            raise PysonnetRuntimeError(
                f"Unsupported operand types for >=: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.AND:
            return left and right
        if operator == ast.Binary.Operator.OR:
            return left or right
        if operator == ast.Binary.Operator.IN:
            if isinstance(right, Array):
                return Boolean(left in right)
            if isinstance(right, dict):
                return Boolean(left in right)
            raise PysonnetRuntimeError(
                f"Unsupported operand types for in: {self._stdlib._type(left)} and {self._stdlib._type(right)}"
            )
        if operator == ast.Binary.Operator.INDEX:
            if isinstance(left, Array) and isinstance(right, Number):
                try:
                    return cast(Primitive, left[right.value])
                except IndexError:
                    raise PysonnetRuntimeError(f"Index out of range, not within [0, {len(left)})")
            if isinstance(left, Object):
                if not isinstance(right, String):
                    raise PysonnetRuntimeError(f"Unsupported type for index: {type(right)}, expected string")
                try:
                    return left[right]
                except KeyError:
                    raise PysonnetRuntimeError(f"Field does not exist: {right}")
            raise PysonnetRuntimeError(f"Unsupported operand types for index: {type(left)} and {type(right)}")
        raise PysonnetRuntimeError(f"Unsupported binary operator: {operator}")

    def _evaluate_apply(self, node: ast.Apply, context: Context) -> Primitive:
        callee = self(node.callee, context)
        if isinstance(callee, Lazy):
            callee = callee()
        if not isinstance(callee, Function):
            raise PysonnetRuntimeError(f"Cannot call {self._stdlib._type(callee)}")
        args = [
            self(arg.expr, context) if node.tailstrict else _make_lazy(self, arg.expr, context)
            for arg in node.args
            if arg.ident is None
        ]
        kwargs = {
            arg.ident.name: self(arg.expr) if node.tailstrict else _make_lazy(self, arg.expr, context)
            for arg in node.args
            if arg.ident is not None
        }
        return cast(Primitive, callee(*args, **kwargs))

    def _evaluate_apply_brace(self, node: ast.ApplyBrace, context: Context) -> Primitive:
        return self._evaluate_binary(ast.Binary(ast.Binary.Operator.ADD, node.left, node.right), context)

    def _evaluate_local(self, node: ast.LocalExpression, context: Context) -> Primitive:
        context = context.clone()
        for bind in node.binds:
            context.bindings[bind.ident.name] = _make_lazy(self, bind.expr, context)
        return self(node.expr, context)

    def _evaluate_if(self, node: ast.IfExpression, context: Context) -> Primitive:
        condition = self(node.condition, context)
        if isinstance(condition, Lazy):
            condition = condition()
        if not isinstance(condition, Boolean):
            raise PysonnetRuntimeError(f"Condition must be a boolean, not {self._stdlib._type(condition)}")
        if condition:
            return self(node.then_expr, context)
        if node.else_expr is not None:
            return self(node.else_expr, context)
        return NULL

    def _evaluate_function(self, node: ast.Function, context: Context) -> Function:
        expr = node.expr
        context = context.clone()

        def _execute_function(*args: Primitive, **kwargs: Primitive) -> Primitive:
            # check parameter names
            unknown_param_names = set(kwargs) - set(param.ident.name for param in node.params)
            if unknown_param_names:
                raise PysonnetRuntimeError(f"Unknown parameters: {', '.join(unknown_param_names)}")
            # check parameter count
            if len(args) + len(kwargs) > len(node.params):
                raise PysonnetRuntimeError(f"Too many parameters, expected {len(node.params)}")
            # check parameter duplication
            args_ = dict(zip((param.ident.name for param in node.params), args))
            duplicate_param_names = set(args_) & set(kwargs)
            if duplicate_param_names:
                raise PysonnetRuntimeError(f"Duplicate parameters: {', '.join(duplicate_param_names)}")
            # merge args into kwargs
            kwargs.update(args_)
            del args
            # fill missing kwargs with default values
            expected_param_names = set(param.ident.name for param in node.params)
            missing_kwparam_names = expected_param_names - set(kwargs)
            default_kwparam_nodes = {
                param.ident.name: param.default for param in node.params if param.default is not None
            }
            if set(missing_kwparam_names) - set(default_kwparam_nodes):
                raise PysonnetRuntimeError(f"Missing parameters: {', '.join(missing_kwparam_names)}")
            default_kwargs = {
                name: _make_lazy(self, default_kwparam_nodes[name], context) for name in missing_kwparam_names
            }
            # update context
            context.bindings.update(kwargs)
            context.bindings.update(default_kwargs)
            return self(expr, context)

        return Function(_execute_function)

    def _evaluate_array_comprehension(self, node: ast.ArrayComprehension, context: Context) -> Array:
        iterable = self(node.forspec.expr, context)
        if isinstance(iterable, Lazy):
            iterable = iterable()
        if not isinstance(iterable, Array):
            raise PysonnetRuntimeError(f"Unexpected type {self._stdlib._type(iterable)}, expected array")
        context = context.clone()
        compspecs = [compspec for compspec in node.compspecs]
        while compspecs and isinstance(compspecs[0], ast.IfSpec):
            ifspec = cast(ast.IfSpec, compspecs.pop(0))
            for index, value in enumerate(iterable):
                context.bindings[node.forspec.ident.name] = value
                condition = self(ifspec.condition, context)
                if isinstance(condition, Lazy):
                    condition = condition()
                if not isinstance(condition, Boolean):
                    raise PysonnetRuntimeError(f"Unpexpected type {self._stdlib._type(condition)}, expected boolean")
                if not condition:
                    iterable.pop(index)
        values: List[Primitive] = []
        for value in iterable:
            context.bindings[node.forspec.ident.name] = value
            if compspecs:
                nested_forspec = compspecs[0]
                assert isinstance(nested_forspec, ast.ForSpec)
                nested_comprehension = ast.ArrayComprehension(node.expression, nested_forspec, compspecs[1:])
                nested_values = self._evaluate_array_comprehension(nested_comprehension, context.clone())
                values.extend(nested_values)
            else:
                values.append(self(node.expression, context))
        return Array(values)

    def _evaluate_object_comprehension(self, node: ast.ObjectComprehension, context: Context) -> Object:
        iterable = self(node.forspec.expr, context)
        if isinstance(iterable, Lazy):
            iterable = iterable()
        if not isinstance(iterable, Array):
            raise PysonnetRuntimeError(f"Unexpected type {self._stdlib._type(iterable)}, expected array")
        context = context.clone()
        for local in node.locals_:
            context.bindings[local.bind.ident.name] = _make_lazy(self, local.bind.expr, context)
        compspecs = [compspec for compspec in node.compspecs]
        while compspecs and isinstance(compspecs[0], ast.IfSpec):
            ifspec = cast(ast.IfSpec, compspecs.pop(0))
            for index, value in enumerate(iterable):
                context.bindings[node.forspec.ident.name] = value
                condition = self(ifspec.condition, context)
                if isinstance(condition, Lazy):
                    condition = condition()
                if not isinstance(condition, Boolean):
                    raise PysonnetRuntimeError(f"Unpexpected type {self._stdlib._type(condition)}, expected boolean")
                if not condition:
                    iterable.pop(index)
        obj = Object()
        context.this = obj
        if context.dollar is None:
            context.dollar = obj
        for value in iterable:
            context.bindings[node.forspec.ident.name] = value

            if compspecs:
                nested_forspec = compspecs[0]
                assert isinstance(nested_forspec, ast.ForSpec)
                nested_comprehension = ast.ObjectComprehension([], node.key, node.value, nested_forspec, compspecs[1:])
                values = self._evaluate_object_comprehension(nested_comprehension, context.clone())
                for key, value in values.items():
                    if key in obj:
                        raise PysonnetRuntimeError(f"Duplicate field name: {key}")
                    obj.add_field(Object.Field(key, value))
            else:
                key_ = self(node.key, context)
                if isinstance(key_, Lazy):
                    key_ = key_()
                if isinstance(key_, Null):
                    continue
                if not isinstance(key_, String):
                    raise PysonnetRuntimeError(f"Field name must be a string, not {self._stdlib._type(key)}")
                key = key_
                if key in obj:
                    raise PysonnetRuntimeError(f"Duplicate field name: {key}")
                value = self(node.value, context)
                obj.add_field(Object.Field(key, value))
        return obj

    def _evaluate_self(self, node: ast.Self, context: Context) -> Primitive:
        if context.this is None:
            raise PysonnetRuntimeError("Can't use self outside of an object.")
        return context.this

    def _evaluate_dollar(self, node: ast.Dollar, context: Context) -> Primitive:
        if context.dollar is None:
            raise PysonnetRuntimeError("No top-level object found.")
        return context.dollar

    def _evaluate_super(self, node: ast.Super, context: Context) -> Primitive:
        if context.super_ is None:
            raise PysonnetRuntimeError("Attempt to use super when there is no super class.")
        return context.super_

    def _evaluate_error(self, node: ast.Error, context: Context) -> Primitive:
        message = self(node.expr, context)
        raise PysonnetRuntimeError(str(message))

    def _evaluate_assert(self, node: ast.AssertExpression, context: Context) -> Primitive:
        condition = self(node.assert_.condition, context)
        if isinstance(condition, Lazy):
            condition = condition()
        if not isinstance(condition, Boolean):
            raise PysonnetRuntimeError(f"Unpexpected type {self._stdlib._type(condition)}, expected boolean")
        if not condition:
            message: Primitive
            if node.assert_.message is None:
                message = String("Assertion failed")
            else:
                message = self(node.assert_.message, context)
            raise PysonnetRuntimeError(str(message))
        return self(node.expr, context)

    def _evaluate_import(self, node: ast.Import, context: Context) -> Primitive:
        del context
        filename = self._rootdir / Path(node.filename)
        if not filename.exists():
            raise PysonnetRuntimeError(f"File not found: {filename}")
        if not filename.is_file():
            raise PysonnetRuntimeError(f"Not a file: {filename}")
        with filename.open() as f:
            jp = Parser(Lexer(f))
            ast = jp.parse()
        if not ast:
            raise PysonnetRuntimeError(f"Failed to parse {filename}")
        evaluator = Evaluator(self._ext_vars, filename)
        return evaluator(ast)

    def _evaluate_importstr(self, node: ast.Importstr, context: Context) -> String:
        del context
        filename = self._rootdir / Path(node.filename)
        if not filename.exists():
            raise PysonnetRuntimeError(f"File not found: {filename}")
        if not filename.is_file():
            raise PysonnetRuntimeError(f"Not a file: {filename}")
        return String(filename.read_text())

    def _evaluate_importbin(self, node: ast.Importbin, context: Context) -> Array[Number[int]]:
        del context
        filename = self._rootdir / Path(node.filename)
        if not filename.exists():
            raise PysonnetRuntimeError(f"File not found: {filename}")
        if not filename.is_file():
            raise PysonnetRuntimeError(f"Not a file: {filename}")
        return Array([Number(b) for b in filename.read_bytes()])

    def __call__(self, node: ast.AST, context: Optional[Context] = None) -> Primitive:
        if context is None:
            context = Context()

        if isinstance(node, ast.LiteralAST):
            return self._evaluate_literal(node, context)

        if isinstance(node, ast.Self):
            return self._evaluate_self(node, context)

        if isinstance(node, ast.Dollar):
            return self._evaluate_dollar(node, context)

        if isinstance(node, ast.Super):
            return self._evaluate_super(node, context)

        if isinstance(node, ast.Object):
            return self._evaluate_object(node, context)

        if isinstance(node, ast.Array):
            return self._evaluate_array(node, context)

        if isinstance(node, ast.Identifier):
            return self._evaluate_identifier(node, context)

        if isinstance(node, ast.Binary):
            return self._evaluate_binary(node, context)

        if isinstance(node, ast.Apply):
            return self._evaluate_apply(node, context)

        if isinstance(node, ast.ApplyBrace):
            return self._evaluate_apply_brace(node, context)

        if isinstance(node, ast.LocalExpression):
            return self._evaluate_local(node, context)

        if isinstance(node, ast.IfExpression):
            return self._evaluate_if(node, context)

        if isinstance(node, ast.Function):
            return self._evaluate_function(node, context)

        if isinstance(node, ast.ArrayComprehension):
            return self._evaluate_array_comprehension(node, context)

        if isinstance(node, ast.ObjectComprehension):
            return self._evaluate_object_comprehension(node, context)

        if isinstance(node, ast.Error):
            return self._evaluate_error(node, context)

        if isinstance(node, ast.AssertExpression):
            return self._evaluate_assert(node, context)

        if isinstance(node, ast.Import):
            return self._evaluate_import(node, context)

        if isinstance(node, ast.Importbin):
            return self._evaluate_importbin(node, context)

        if isinstance(node, ast.Importstr):
            return self._evaluate_importstr(node, context)

        raise PysonnetRuntimeError(f"Unsupported type: {type(node)}")



import argparse
import json
import os
import sys
from io import StringIO
from pathlib import Path
from typing import Dict, List, Optional, TextIO

from pysonnet import __version__
from pysonnet.ast import asdict
from pysonnet.errors import PysonnetRuntimeError
from pysonnet.evaluator import Evaluator
from pysonnet.lexer import Lexer
from pysonnet.parser import Parser


def _parse_ext_vars(inputs: List[str]) -> Dict[str, str]:
    ext_vars: Dict[str, str] = {}
    for variable in inputs:
        if "=" not in variable:
            name = variable
            value = os.getenv(variable)
            if value is None:
                print(f"Environment variable {variable} was undefined.", file=sys.stderr)
                sys.exit(1)
        else:
            name, value = variable.split("=", 1)
        ext_vars[name] = value
    return ext_vars


def _show_errors(errors: List[str]) -> None:
    for error in errors:
        print(error, file=sys.stderr)


def main(prog: Optional[str] = None) -> None:
    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument("input", nargs="?", type=str, help="input file")
    parser.add_argument("-e", "--exec", action="store_true")
    parser.add_argument("-V", "--ext-str", type=str, action="append", default=[], help="external string variable")
    parser.add_argument("--ast", action="store_true", help="show the abstract syntax tree")
    parser.add_argument("--indent", type=int, default=None, help="indentation level for JSON output")
    parser.add_argument("--ensure-ascii", action="store_true", help="ensure ASCII output")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args()

    textio: TextIO
    filename: Optional[Path] = None
    if args.input is None:
        if not sys.stdin.isatty():
            textio = StringIO(sys.stdin.read())
        else:
            parser.error("the following arguments are required: input")
    else:
        if args.exec:
            textio = StringIO(args.input)
        else:
            textio = open(args.input)
            filename = Path(args.input)

    with textio:
        jp = Parser(Lexer(textio))
        ast = jp.parse()

    if not ast:
        _show_errors(jp.errors)
        sys.exit(1)

    if args.ast:
        print(json.dumps(asdict(ast), indent=args.indent, ensure_ascii=args.ensure_ascii))

    ext_vars = _parse_ext_vars(args.ext_str)

    try:
        evaluator = Evaluator(ext_vars, filename)
        value = evaluator(ast)
    except PysonnetRuntimeError as e:
        print("Runtime Error:", e.args[0], file=sys.stderr)
        sys.exit(1)

    print(json.dumps(value.to_json(), indent=args.indent, ensure_ascii=args.ensure_ascii))



import json
from importlib.metadata import version
from io import StringIO
from os import PathLike
from typing import Mapping, Optional, Union

from pysonnet.errors import PysonnetSyntaxError
from pysonnet.evaluator import Evaluator
from pysonnet.lexer import Lexer
from pysonnet.parser import Parser
from pysonnet.types import JsonPrimitive

__version__ = version("pysonnet")
__all__ = ["Evaluator", "Lexer", "Parser", "evaluate_file", "load", "loads"]


def load(
    filename: Union[str, PathLike],
    ext_vars: Optional[Mapping[str, str]] = None,
    *,
    encoding: Optional[str] = None,
) -> JsonPrimitive:
    with open(filename, "r", encoding=encoding) as jsonnetfile:
        lexer = Lexer(jsonnetfile)
        parser = Parser(lexer)
        node = parser.parse()
        if node is None:
            raise PysonnetSyntaxError(*parser.errors)
        evaluator = Evaluator(ext_vars, filename)
        value = evaluator(node)
    return value.to_json()


def loads(s: str, ext_vars: Optional[Mapping[str, str]] = None) -> JsonPrimitive:
    lexer = Lexer(StringIO(s))
    parser = Parser(lexer)
    node = parser.parse()
    if node is None:
        raise PysonnetSyntaxError(*parser.errors)
    evaluator = Evaluator(ext_vars)
    value = evaluator(node)
    return value.to_json()


def evaluate_file(
    filename: Union[str, PathLike],
    ext_vars: Optional[Mapping[str, str]] = None,
    *,
    encoding: Optional[str] = None,
    indent: Optional[int] = None,
    ensure_ascii: bool = True,
) -> str:
    return json.dumps(
        load(filename, ext_vars, encoding=encoding),
        indent=indent,
        ensure_ascii=ensure_ascii,
    )



import abc
import dataclasses
import enum
from typing import Any, Callable, Dict, Generic, List, Literal, Optional, TypeVar, Union

_PRIMITIVE = Union[str, int, float, bool, None, dict, list, Callable]
_T = TypeVar("_T", bound=_PRIMITIVE)
_S = TypeVar("_S", bound=_PRIMITIVE)
_T_co = TypeVar("_T_co", covariant=True, bound=_PRIMITIVE)
_S_co = TypeVar("_S_co", covariant=True, bound=_PRIMITIVE)


@dataclasses.dataclass(frozen=True)
class AST(abc.ABC, Generic[_T_co]):
    ...


@dataclasses.dataclass(frozen=True)
class LiteralAST(AST[_T_co]):
    value: _T_co


@dataclasses.dataclass(frozen=True)
class Identifier(Generic[_T_co], AST[_T_co]):
    name: str


@dataclasses.dataclass(frozen=True)
class Null(LiteralAST[Literal[None]]):
    value: Literal[None] = None


@dataclasses.dataclass(frozen=True)
class Boolean(LiteralAST[bool]):
    ...


@dataclasses.dataclass(frozen=True)
class Number(LiteralAST[Union[int, float]]):
    ...


@dataclasses.dataclass(frozen=True)
class String(LiteralAST[str]):
    ...


@dataclasses.dataclass(frozen=True)
class Error(Generic[_T_co], AST[_T_co]):
    expr: AST[_T_co]


@dataclasses.dataclass(frozen=True)
class Assert:
    condition: AST[bool]
    message: Optional[AST[Any]] = None


@dataclasses.dataclass(frozen=True)
class IfExpression(Generic[_T_co, _S_co], AST[Union[_T_co, _S_co, None]]):
    condition: AST[bool]
    then_expr: AST[_T_co]
    else_expr: Optional[AST[_S_co]] = None


@dataclasses.dataclass(frozen=True)
class ForSpec(Generic[_T_co]):
    ident: Identifier[_T_co]
    expr: AST[_T_co]


@dataclasses.dataclass(frozen=True)
class IfSpec:
    condition: AST[bool]


ComprehensionSpec = Union[ForSpec, IfSpec]


@dataclasses.dataclass(frozen=True)
class Bind(Generic[_T]):
    ident: Identifier
    expr: AST[_T]


@dataclasses.dataclass(frozen=True)
class LocalExpression(Generic[_T_co], AST[_T_co]):
    binds: List[Bind]
    expr: AST[_T_co]


@dataclasses.dataclass(frozen=True)
class AssertExpression(Generic[_T_co], AST[_T_co]):
    assert_: Assert
    expr: AST[_T_co]


@dataclasses.dataclass(frozen=True)
class Unary(Generic[_T_co], AST[_T_co]):
    class Operator(str, enum.Enum):
        PLUS = "PLUS"
        MINUS = "MINUS"
        NOT = "NOT"
        BITWISE_NOT = "BITWISE_NOT"

    operator: Operator
    operand: AST[_T_co]


@dataclasses.dataclass(frozen=True)
class Binary(Generic[_T_co], AST[_T_co]):
    class Operator(str, enum.Enum):
        ADD = "ADD"
        SUB = "SUB"
        MUL = "MUL"
        DIV = "DIV"
        MOD = "MOD"
        BITWISE_AND = "BITWISE_AND"
        BITWISE_OR = "BITWISE_OR"
        BITWISE_XOR = "BITWISE_XOR"
        LSHIFT = "LSHIFT"
        RSHIFT = "RSHIFT"
        AND = "AND"
        OR = "OR"
        EQ = "EQ"
        NE = "NE"
        LT = "LT"
        LE = "LE"
        GT = "GT"
        GE = "GE"
        IN = "IN"
        INDEX = "INDEX"

    operator: Operator
    left: AST[Any]
    right: AST[Any]


@dataclasses.dataclass(frozen=True)
class Self(AST[Dict[str, Any]]):
    ...


@dataclasses.dataclass(frozen=True)
class Dollar(AST[Dict[str, Any]]):
    ...


@dataclasses.dataclass(frozen=True)
class Super(AST[Dict[str, Any]]):
    ...


@dataclasses.dataclass(frozen=True)
class Param(Generic[_T_co]):
    ident: Identifier
    default: Optional[AST[_T_co]] = None


@dataclasses.dataclass(frozen=True)
class Function(Generic[_T_co], AST[Callable[..., _T_co]]):
    params: List[Param]
    expr: AST[_T_co]


@dataclasses.dataclass(frozen=True)
class Arg(Generic[_T_co]):
    expr: AST[_T_co]
    ident: Optional[Identifier] = None


@dataclasses.dataclass(frozen=True)
class Apply(Generic[_T_co], AST[_T_co]):
    callee: AST[Callable[..., _T_co]]
    args: List[Arg[Any]]
    tailstrict: bool = False


@dataclasses.dataclass(frozen=True)
class ApplyBrace(AST[Dict[str, Any]]):
    left: AST[Dict[str, Any]]
    right: AST[Dict[str, Any]]


@dataclasses.dataclass(frozen=True)
class ObjectLocal(Generic[_T_co]):
    bind: Bind[_T_co]


@dataclasses.dataclass(frozen=True)
class ObjectField(Generic[_T_co]):
    class Visibility(str, enum.Enum):
        VISIBLE = "VISIBLE"
        HIDDEN = "HIDDEN"
        FORCE_VISIBLE = "FORCE_VISIBLE"

    key: AST[Union[str, None]]
    expr: AST[_T_co]
    inherit: bool = False
    visibility: Visibility = Visibility.VISIBLE


ObjectMember = Union[ObjectField, ObjectLocal, Assert]


@dataclasses.dataclass(frozen=True)
class Object(AST[Dict[str, Any]]):
    members: List[ObjectMember]


@dataclasses.dataclass(frozen=True)
class ObjectComprehension(Generic[_T_co], AST[Dict[str, _T_co]]):
    locals_: List[ObjectLocal]
    key: AST[Optional[str]]
    value: AST[_T_co]
    forspec: ForSpec[Any]
    compspecs: List[ComprehensionSpec]


@dataclasses.dataclass(frozen=True)
class Array(Generic[_T_co], AST[List[_T_co]]):
    elements: List[AST[_T_co]]


@dataclasses.dataclass(frozen=True)
class ArrayComprehension(Generic[_T_co], AST[List[_T_co]]):
    expression: AST[Any]
    forspec: ForSpec[_T_co]
    compspecs: List[ComprehensionSpec]


@dataclasses.dataclass(frozen=True)
class Import(AST[Any]):
    filename: str


@dataclasses.dataclass(frozen=True)
class Importstr(AST[str]):
    filename: str


@dataclasses.dataclass(frozen=True)
class Importbin(AST[List[int]]):
    filename: str


def asdict(ast: AST) -> Dict[str, Any]:
    d: Dict[str, Any] = {"__AST__": ast.__class__.__name__}
    for field in dataclasses.fields(ast):
        value = getattr(ast, field.name)
        if dataclasses.is_dataclass(value):
            d[field.name] = asdict(value)
        elif isinstance(value, list):
            d[field.name] = [asdict(x) if dataclasses.is_dataclass(x) else x for x in value]
        elif isinstance(value, dict):
            d[field.name] = {k: asdict(v) if dataclasses.is_dataclass(v) else v for k, v in value.items()}
        else:
            d[field.name] = value
    return d
